package net.sf.basedb.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.StatelessSession;
import org.hibernate.dialect.Dialect;
import org.hibernate.jdbc.Work;
import org.hibernate.query.Query;

import net.sf.basedb.core.AnnotationBatcher.InsertBatcher;
import net.sf.basedb.core.AnnotationBatcher.UpdateBatcher;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;
import net.sf.basedb.core.log.LoggingInterceptor;
import net.sf.basedb.core.log.db.DbLogManager;
import net.sf.basedb.core.log.db.DbLogManagerFactory;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.snapshot.SnapshotManager;

public class MigrateRbaToDba 
	implements Work
{

	public static void migrateAll(SessionControl sc, File cfg, boolean dryRun, ProgressReporter progress)
		throws SQLException, IOException
	{
		DbControl dc = null;
		try
		{
			if (dryRun)
			{
				progress.display(0, "Dry-run, changes will not be permanent\n");
			}
			SubtypeMapping mapping = new SubtypeMapping(cfg);
			progress.display(2, "Loaded "+mapping.lines.size()+ " mappings from "+cfg.getName()+"\n");

			dc = sc.newDbControl("Migrate raw bioassays to derived bioassays");
			mapping.validateMapping(dc);
			
			// Register a special logger for the migrated information
			DbLogManagerFactory logFactory = DbLogManagerFactory.getInstanceIfEnabled(dc.getSessionControl());
			if (logFactory != null)
			{
				logFactory.setSpecialLogger(Rba.class, new MigratedLogger());
			}
			
			MigrateRbaToDba migrater = new MigrateRbaToDba(dc, mapping, progress);
			HibernateUtil.doWork(dc.getHibernateSession(), migrater);
		
			if (dryRun)
			{
				progress.display(98, "Undoing all changes...\n");
				dc.close();
				progress.display(100, "Dry-run completed successfully!\n");
			}
			else
			{
				progress.display(98, "Committing transaction...\n");
				dc.commit();
				progress.display(100, "Migration completed successfully!\n");
			}
			
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	private final DbControl dc;
	private final LoggingInterceptor logInterceptor;
	private final SubtypeMapping mapping;
	private final ProgressReporter progress;
	
	private final StatelessSession session;
	private final Dialect dialect;
	
	// All batchers (to make it easy to flush/close all at once)
	private final List<InsertBatcher> allInsertBatchers;
	private final List<UpdateBatcher> allUpdateBatchers;
	
	private MigrateRbaToDba(DbControl dc, SubtypeMapping mapping, ProgressReporter progress)
	{
		this.dc = dc;
		this.logInterceptor = dc.getLoggingInterceptor();
		this.mapping = mapping;
		this.progress = progress;
		
		this.session = dc.getStatelessSession();
		this.dialect = HibernateUtil.getDialect();
		
		this.allInsertBatchers = new ArrayList<InsertBatcher>();
		this.allUpdateBatchers = new ArrayList<UpdateBatcher>();
	}
	
	/*
		From Work interface
		-------------------
	*/
	@Override
	public void execute(Connection connection)
		throws SQLException
	{
		try
		{
			// Create batchers
			InsertBatcher insertDerivedBioAssay = createInsertBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_INSERT_INTO_DERIVEDBIOASSAYS"),
				Types.INTEGER, /* version */
				Types.INTEGER, /* fileset_id */
				Types.DATE,    /* entry_date */
				Types.INTEGER, /* subtype_id */
				Types.INTEGER, /* job_id */
				Types.INTEGER, /* software_id */
				Types.INTEGER, /* protocol_id */
				Types.INTEGER, /* extract_id */
				Types.INTEGER, /* annotationset_id */
				Types.VARCHAR, /* name */
				Types.LONGVARCHAR, /* description */
				Types.INTEGER, /* removed_by */
				Types.INTEGER, /* itemkey_id */
				Types.INTEGER, /* projectkey_id */
				Types.INTEGER  /* owner */
				);
			
			UpdateBatcher linkParentDba = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_PARENT_DERIVEDBIOASSAY"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* parentId */
				);
			
			UpdateBatcher linkParentPba = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_PARENT_PHYSICALBIOASSAYS"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* parentId */
				);
			
			InsertBatcher linkArrayDesign = createInsertBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_ARRAYDESIGN"),
				Types.VARCHAR, /* name */
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* arrayDesignId */
				);
			
			UpdateBatcher cleanRawBioAssay = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_CLEAN_RAWBIOASSAY"),
				Types.INTEGER /* rbaId */
				);
			
			UpdateBatcher linkAnnotationSet = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_ANNOTATIONSET"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* annotationSetId */
				);
			
			UpdateBatcher linkFileSet = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_FILESET"),
				Types.INTEGER /* fileSetId */
				);
			
			UpdateBatcher updateFileSetMember = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_UPDATE_MEMBERTYPE"),
				Types.INTEGER, /* newFileTypeId */
				Types.INTEGER, /* currentFileTypeId */
				Types.INTEGER  /* fileSetId */
				);
			
			UpdateBatcher linkAnyToAnyFrom = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_ANYTOANY_FROM"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* rbaId */
				);
			
			UpdateBatcher linkAnyToAnyTo = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_ANYTOANY_TO"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* rbaId */
				);
			
			UpdateBatcher linkChangeHistory = createUpdateBatcher(connection, 
				PredefinedQuery.getQueryString("RBA2DBA_LINK_CHANGE_HISTORY"),
				Types.INTEGER, /* dbaId */
				Types.INTEGER  /* rbaId */
				);
			
			// Count existing raw bioassays -- seems like SQL returns a BigInteger (and not a Long)
			// as Hiberate does so we use Number to be on the safe side
			Query<Number> countQuery = HibernateUtil.getPredefinedSQLQuery(session, "RBA2DBA_COUNT_RAWBIOASSAYS", Number.class);
			long numRba = HibernateUtil.loadData(countQuery).longValue();
			progress.display(5, "Starting migration of " + numRba + " raw bioassays...\n");
			
			// Load all raw bioassays
			Query<Object[]> rbaQuery = HibernateUtil.getPredefinedSQLQuery(session, "RBA2DBA_LOAD_RAWBIOASSAYS", Object[].class);
			Iterator<Object[]> it = HibernateUtil.loadIterator(rbaQuery, true, dc.getSessionControl());
		
			long nextProgress = System.currentTimeMillis()+1000;
			long numMigrated = 0;
			List<Integer> allAnnotationSets = new ArrayList<>();
			List<Integer> allFileSets = new ArrayList<>();
			Map<Integer, Integer> rbaToDba = new HashMap<>();
			while (it.hasNext())
			{
				if (System.currentTimeMillis() > nextProgress)
				{
					ThreadSignalHandler.checkInterrupted();
					nextProgress = System.currentTimeMillis()+1000;
					progress.display(10+(int)((numMigrated*70)/numRba), 
						"Migrated "+numMigrated+" raw bioassays so far...");
				}
				Rba rba = new Rba(it.next());
				mapping.mapToSubtype(rba);
				rba.createDerivedBioAssay(insertDerivedBioAssay);
				if (rba.bioAssayId != null)
				{
					rba.linkParentBioAssays(linkParentDba, linkParentPba);
				}
				if (rba.arrayDesignId != null)
				{
					rba.linkArrayDesign(linkArrayDesign);
				}
				rba.clean(cleanRawBioAssay);
				if (rba.annotationSetId != null)
				{
					rba.linkAnnotations(linkAnnotationSet);
					allAnnotationSets.add(rba.annotationSetId);
				}
				if (rba.fileSetId != null)
				{
					rba.linkFiles(linkFileSet, updateFileSetMember);
					allFileSets.add(rba.fileSetId);
				}
				rba.linkAnyToAny(linkAnyToAnyFrom, linkAnyToAnyTo);
				
				rba.linkChangeHistory(linkChangeHistory);
				if (logInterceptor != null)
				{
					// NOTE! It is important that we flush before adding the new log
					// entries or they will both end up on the new derived bioassay
					linkChangeHistory.flush();
					logInterceptor.addManualSingleChange(ChangeType.UPDATE, rba);
				}

				rbaToDba.put(rba.rbaId, rba.dbaId);
				numMigrated++;
				if (numMigrated % 100 == 0) flush();
			}
			flush();
			progress.display(80, "Successfully migrated "+numMigrated+" raw bioassays\n");

			progress.display(80, "Item lists...");
			updateItemLists(connection, rbaToDba);

			progress.display(90, "Job parameters...");
			linkItemValues(connection, rbaToDba);

			progress.display(92, "Annotation types...");
			enableAnnotationTypes(connection, allAnnotationSets);
			
			progress.display(94, "File types...");
			enableFileTypes(connection, allFileSets);
			
			flush(); // Should not be needed, but just in case...
			progress.display(95, "Cached items...");
			SnapshotManager.removeSnapshots(allAnnotationSets);
			progress.display(96, "Cached items ("+allAnnotationSets.size()+")\n");
		}
		finally
		{
			for (UpdateBatcher b : allUpdateBatchers)
			{
				b.close();
			}
			for (InsertBatcher b : allInsertBatchers)
			{
				b.close();
			}
		}
	}
	// -------------------

	/**
		Make sure that the annotation types that are used by the migrated raw
		bioassays are enabled for derived bioassays.
	*/
	private void enableAnnotationTypes(Connection connection, List<Integer> allAnnotationSets)
		throws SQLException
	{
		int maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();
		// Make sure that all annotation types are enabled for derived bioassays
		Query<Integer> getAnnotationTypes = (Query<Integer>)HibernateUtil.getPredefinedSQLQuery(session, 
				"RBA2DBA_LOAD_ANNOTATIONTYPE_IDS", Integer.class);
		Set<Integer> annotationTypes = new HashSet<>();
		int startIndex = 0;
		int endIndex = Math.min(maxParameters, allAnnotationSets.size());
		while (startIndex < allAnnotationSets.size())
		{
			List<Integer> sublist = allAnnotationSets.subList(startIndex, endIndex);
			getAnnotationTypes.setParameterList("allAnnotationSets", sublist, Type.INT.getTypeWrapper().getHibernateType());
			annotationTypes.addAll(getAnnotationTypes.list());
			startIndex = endIndex;
			endIndex = Math.min(startIndex + maxParameters, allAnnotationSets.size());
		}
		UpdateBatcher insertAnnotationType = createUpdateBatcher(connection, 
			PredefinedQuery.getQueryString("RBA2DBA_INSERT_INTO_ANNOTATIONTYPEITEMS"),
			Types.INTEGER, /* annotationtype_id */
			Types.INTEGER /* item_type */
		);
		for (Integer atId : annotationTypes)
		{
			insertAnnotationType.addToBatch(atId, 268);
		}
		insertAnnotationType.flush();
		progress.display(93, "Annotation types ("+annotationTypes.size()+")\n");
	}
	
	/**
		Make sure that the data file types that are used by the migrated raw
		bioassays are enabled for derived bioassays.
	*/
	private void enableFileTypes(Connection connection, List<Integer> allFileSets)
		throws SQLException
	{
		int maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();
		// Make sure that all annotation types are enabled for derived bioassays
		Query<Integer> getFileTypes = (Query<Integer>)HibernateUtil.getPredefinedSQLQuery(session, 
				"RBA2DBA_LOAD_FILETYPES", Integer.class);
		Set<Integer> fileTypes = new HashSet<>();
		int startIndex = 0;
		int endIndex = Math.min(maxParameters, allFileSets.size());
		while (startIndex < allFileSets.size())
		{
			List<Integer> sublist = allFileSets.subList(startIndex, endIndex);
			getFileTypes.setParameterList("allFileSets", sublist, Type.INT.getTypeWrapper().getHibernateType());
			fileTypes.addAll(getFileTypes.list());
			startIndex = endIndex;
			endIndex = Math.min(startIndex + maxParameters, allFileSets.size());
		}
		UpdateBatcher updateFileType = createUpdateBatcher(connection, 
			PredefinedQuery.getQueryString("RBA2DBA_UPDATE_FILETYPE"),
			Types.INTEGER /* id */
		);
		for (Integer ftId : fileTypes)
		{
			updateFileType.addToBatch(ftId);
		}
		updateFileType.flush();
		progress.display(94, "File types ("+fileTypes.size()+")\n");
	}
	
	/**
		Change raw bioassays that have been used as job parameters to the
		new derived bioassay. It would have been possible to do this 
		per raw bioassay but it goes a lot faster if we can use the 'id'
		column in the update query.
	*/
	private void linkItemValues(Connection connection, Map<Integer, Integer> rbaToDba)
		throws SQLException
	{
		UpdateBatcher linkItemValues = createUpdateBatcher(connection, 
			PredefinedQuery.getQueryString("RBA2DBA_LINK_ITEMVALUES"),
			Types.INTEGER, /* dbaId */
			Types.INTEGER, /* rbaId */
			Types.INTEGER  /* valueId */
			);
		Query<Object[]> itemValueQuery = HibernateUtil.getPredefinedSQLQuery(session, "RBA2DBA_LOAD_ITEMVALUES", Object[].class);
		List<Object[]> list = HibernateUtil.loadList(itemValueQuery, dc.getSessionControl());
		int numValues = 0;
		for (Object[] data : list)
		{
			Integer valueId = (Integer)data[0];
			Integer rbaId = (Integer)data[1];
			Integer dbaId = rbaToDba.get(rbaId);
			numValues++;
			if (dbaId != null)
			{
				// Some raw bioassays may no longer exist but the
				// parameters are not deleted so we should not try
				// to make any changes to them
				linkItemValues.addToBatch(dbaId, rbaId, valueId);
			}
			if (numValues % 1000 == 0)
			{
				progress.display(91, "Job parameters ("+numValues+" of "+list.size()+")...");
				linkItemValues.flush();
			}
		}
		linkItemValues.flush();
		progress.display(92, "Job parameters ("+numValues+")                             \n");
	}
	
	/**
		Update item lists with raw bioassays to item lists with derived 
		bioassays.
	*/
	private void updateItemLists(Connection connection, Map<Integer, Integer> rbaToDba)
		throws SQLException
	{
		
		Query<Object[]> listQuery = HibernateUtil.getPredefinedSQLQuery(session, 
			"RBA2DBA_LOAD_ITEMLISTS", Object[].class);
		Map<Integer, Itemlist> lists = new HashMap<>();
		for (Object[] data : HibernateUtil.loadList(listQuery, null))
		{
			Itemlist list = new Itemlist(data);
			mapping.mapToSubtype(list);
			lists.put(list.listId, list);
		}
		
		UpdateBatcher moveListMember = createUpdateBatcher(connection, 
			PredefinedQuery.getQueryString("RBA2DBA_MOVE_LISTMEMBER"),
			Types.INTEGER, /* dbaId */
			Types.INTEGER, /* rbaId */
			Types.INTEGER  /* listId */
			);
		UpdateBatcher deleteListMember = createUpdateBatcher(connection,
			PredefinedQuery.getQueryString("RBA2DBA_DELETE_LISTMEMBER"),
			Types.INTEGER, /* rbaId */
			Types.INTEGER  /* listId */
			);
		Query<Object[]> memberQuery = HibernateUtil.getPredefinedSQLQuery(session, 
			"RBA2DBA_LOAD_ITEMLISTMEMBERS", Object[].class);
		List<Object[]> members = HibernateUtil.loadList(memberQuery, null);
		
		int numMembers = 0;
		int totalMembers = members.size();
		int numDeleted = 0;
		for (Object[] data : members)
		{
			Integer listId = (Integer)data[0];
			Integer rbaId = (Integer)data[1];
			Integer dbaId = rbaToDba.get(rbaId);
			Itemlist list = lists.get(listId);
			if (dbaId != null && list != null)
			{
				moveListMember.addToBatch(dbaId, rbaId, listId);
				list.size++;
			}
			else
			{
				// delete entries that point to non-existing items
				deleteListMember.addToBatch(rbaId, listId);
				numDeleted++;
			}
			numMembers++;
			if (numMembers % 1000 == 0)
			{
				progress.display(80+(numMembers*10)/totalMembers, "Item lists ("+numMembers+" of "+totalMembers+")...");
				moveListMember.flush();
			}
		}
		moveListMember.flush();
		
		UpdateBatcher updateItemList = createUpdateBatcher(connection,
			PredefinedQuery.getQueryString("RBA2DBA_UPDATE_ITEMLIST"),
			Types.INTEGER, /* subtypeId */
			Types.INTEGER, /* size */
			Types.INTEGER  /* listId */
			);
		for (Itemlist list : lists.values())
		{
			list.updateEntry(updateItemList);
		}
		updateItemList.flush();
		
		Query<?> markSyncFilters = HibernateUtil.getPredefinedSQLQuery(session, 
				"RBA2DBA_MARK_ITEMLISTFILTERS");
		HibernateUtil.executeUpdate(markSyncFilters);

		Query<?> markItemLists = HibernateUtil.getPredefinedSQLQuery(session, 
			"RBA2DBA_MARK_ITEMLISTS");
		HibernateUtil.executeUpdate(markItemLists);
		
		progress.display(90, "Item lists ("+numMembers+")                                    \n");
	}
	
	private InsertBatcher createInsertBatcher(Connection c, String sql, int... types)
		throws SQLException
	{
		sql = sql.replace('[', dialect.openQuote()).replace(']', dialect.closeQuote());
		InsertBatcher b = new InsertBatcher(c, sql, types);
		allInsertBatchers.add(b);
		return b;
	}

	private UpdateBatcher createUpdateBatcher(Connection c, String sql, int... types)
		throws SQLException
	{
		sql = sql.replace('[', dialect.openQuote()).replace(']', dialect.closeQuote());
		UpdateBatcher b = new UpdateBatcher(c, sql, types);
		allUpdateBatchers.add(b);
		return b;
	}

	/**
		Flush all batchers
	*/
	private void flush()
	{
		String currentSql = null;
		try
		{
			for (UpdateBatcher b : allUpdateBatchers)
			{
				currentSql = b.getSql();
				b.flush();
			}
		}
		catch (SQLException ex)
		{
			throw new BaseException(currentSql, ex);
		}
	}

	
	/**
		Hold information a raw bioassay.
	*/
	public static class Rba
		implements LoggableData
	{
		final Integer rbaId, version;
		final Date entryDate;
		final Integer fileSetId, jobId, softwareId, protocolId;
		final Integer arrayDesignId, bioAssayId, extractId, annotationSetId;
		final String name, description;
		final Integer removedBy, itemKeyId, projectKeyId;
		final Integer ownerId, platformId, variantId;
		final String rawDataType;
		
		// new values after migration
		int dbaId;
		Integer subtypeId;
		Integer arrayDesignLinkId;
		Map<Integer, Integer> fileTypes;
		
		Rba(Object[] data)
		{
			int col = 0;
			rbaId = (Integer)data[col++];
			version = (Integer)data[col++];
			fileSetId = (Integer)data[col++];
			entryDate = (Date)data[col++];
			jobId = (Integer)data[col++];
			softwareId = (Integer)data[col++];
			protocolId = (Integer)data[col++];
			arrayDesignId = (Integer)data[col++];
			bioAssayId = (Integer)data[col++];
			extractId = (Integer)data[col++];
			annotationSetId = (Integer)data[col++];
			name = (String)data[col++];
			description = (String)data[col++];
			removedBy = (Integer)data[col++];
			itemKeyId = (Integer)data[col++];
			projectKeyId = (Integer)data[col++];
			ownerId = (Integer)data[col++];
			platformId = (Integer)data[col++];
			variantId = (Integer)data[col++];
			rawDataType = (String)data[col++];
		}
		
		@Override
		public String toString()
		{
			return "Migrating "+name+": "+rbaId+"->"+dbaId;
		}
		
		/**
			Create the new derived bioassay.
		*/
		void createDerivedBioAssay(InsertBatcher batcher)
			throws SQLException
		{
			dbaId = batcher.insert(
				version, fileSetId, entryDate,
				subtypeId, jobId, softwareId, protocolId,
				extractId, annotationSetId, name, description,
				removedBy, itemKeyId, projectKeyId, ownerId
				);
		}
		
		void linkParentBioAssays(UpdateBatcher batcherDba, 
			UpdateBatcher batcherPba)
			throws SQLException
		{
			batcherDba.addToBatch(dbaId, bioAssayId);
			batcherPba.addToBatch(dbaId, bioAssayId);
		}
		
		void linkArrayDesign(InsertBatcher batcher)
			throws SQLException
		{
			arrayDesignLinkId = batcher.insert("ArrayDesign", dbaId, arrayDesignId);
		}
		
		/**
			Remove links to from the raw bioassay to
			annotations, files, etc.
		*/
		void clean(UpdateBatcher batcher)
			throws SQLException
		{
			batcher.addToBatch(rbaId);
		}
		
		/**
			Link the annotation set to the new derived bioassay.
		*/
		void linkAnnotations(UpdateBatcher batcher)
			throws SQLException
		{
			batcher.addToBatch(dbaId, annotationSetId);
		}
		
		/**
			Link the file set to the new derived bioassay and
			update file types on members.
		*/
		void linkFiles(UpdateBatcher batcher, UpdateBatcher updateMember)
			throws SQLException
		{
			batcher.addToBatch(fileSetId);
			if (fileTypes != null)
			{
				for (Map.Entry<Integer, Integer> entry : fileTypes.entrySet())
				{
					updateMember.addToBatch(entry.getValue(), entry.getKey(), fileSetId);
				}
			}
		}
		
		/**
			Link any-to-any links with the new derived bioassay.
		*/
		void linkAnyToAny(UpdateBatcher from, UpdateBatcher to)
			throws SQLException
		{
			from.addToBatch(dbaId, rbaId);
			to.addToBatch(dbaId, rbaId);
		}
		
		/**
			Link change history entries with the new derived bioassay.
			We need to the session id to avoid moving the newly
			inserted entries about the migration.
		*/
		void linkChangeHistory(UpdateBatcher batcher)
			throws SQLException
		{
			batcher.addToBatch(dbaId, rbaId);
		}

	}
	
	public static class Itemlist
	{
		final Integer listId;
		final String rawDataType;
		final String name;
		
		Integer subtypeId;
		int size;
		
		Itemlist(Object[] data)
		{
			int col = 0;
			listId = (Integer)data[col++];
			rawDataType = (String)data[col++];
			name = (String)data[col++];
		}
		
		/**
			Update the item list entry.
		*/
		void updateEntry(UpdateBatcher batcher)
			throws SQLException
		{
			batcher.addToBatch(subtypeId, size, listId);
		}

	}
	
	/**
		Hold the mappings defined in the mapping file.
	*/
	public static class SubtypeMapping
	{
		
		final File cfg;
		final List<LineMapping> lines;
		
		public SubtypeMapping(File cfg)
			throws IOException
		{
			this.cfg = cfg;
			this.lines = new ArrayList<>();
			
			BufferedReader reader = new BufferedReader(new FileReader(cfg));
			int lineNo = 0;
			do
			{
				String line = reader.readLine();
				if (line == null) break;
				line = line.strip();
				lineNo++;
				if (line.length() > 0 && !line.startsWith("#"))
				{
					lines.add(new LineMapping(lineNo, line));
				}
			}
			while (true);
			reader.close();
		}
		
		void validateMapping(DbControl dc)
		{
			List<String> errors = new ArrayList<>();
			for (LineMapping line : lines)
			{
				line.validateMapping(dc, errors);
			}
			if (errors.size() > 0)
			{
				throw new ConfigurationException(errors.size()+" error(s) in the configuration:\n"+
					net.sf.basedb.util.Values.getString(errors, "\n", true));
			}
		}
		
		/**
			Map a subtype to the given raw bioassay.
		*/
		boolean mapToSubtype(Rba rba)
		{
			for (LineMapping line : lines)
			{
				if (line.matches(rba))
				{
					rba.subtypeId = line.subtypeId;
					rba.fileTypes = line.fileTypes;
					return true;
				}
			}
			rba.subtypeId = null;
			return false;
		}
		
		/**
			Map a subtype to the given item list.
		*/
		boolean mapToSubtype(Itemlist list)
		{
			for (LineMapping line : lines)
			{
				if (line.matches(list))
				{
					list.subtypeId = line.subtypeId;
					return true;
				}
			}
			list.subtypeId = null;
			return false;
		}
		
	}
	
	public static class LineMapping
	{
		final int lineNo;
		final String line;
		final String rawDataType;
		final Integer platformId;
		final Integer variantId;
		final Integer subtypeId;
		final Map<Integer, Integer> fileTypes;
		final String comment;
		
		public LineMapping(int lineNo, String line)
		{
			this.lineNo = lineNo;
			this.line = line;
			String[] cols = line.split("\t", 6);
			if (cols.length < 5)
			{
				throw new ConfigurationException("[Line "+lineNo+"] Need at least 5 columns: "+line);
			}
			else
			{
				rawDataType = net.sf.basedb.util.Values.getStringOrNull(cols[0]);
				platformId = net.sf.basedb.util.Values.getInteger(cols[1], null);
				variantId = net.sf.basedb.util.Values.getInteger(cols[2], null);
				subtypeId = net.sf.basedb.util.Values.getInteger(cols[3], null);
				fileTypes = new HashMap<>();
				String[] maps = cols[4].strip().split(",");
				for (String m : maps)
				{
					if (m.length() > 0)
					{
						String[] kv = m.split("=");
						if (kv.length != 2)
						{
							throw new ConfigurationException("[Line "+lineNo+"] Invalid key=value mapping: "+cols[4]);
						}
						Integer fromId = net.sf.basedb.util.Values.getInteger(kv[0], null);
						Integer toId = net.sf.basedb.util.Values.getInteger(kv[1], null);
						if (fromId == null || toId == null)
						{
							throw new ConfigurationException("[Line "+lineNo+"] Invalid key=value mapping: "+cols[4]);
						}
						fileTypes.put(fromId, toId);
					}
				}
				comment = cols.length < 6 ? null : cols[5];
			}
		}
		
		void validateMapping(DbControl dc, List<String> errors)
		{
			if (rawDataType != null && RawDataTypes.getRawDataType(rawDataType) == null)
			{
				errors.add("[Line "+lineNo+"] Unknown raw data type: "+rawDataType);
			}
			if (itemIsMissing(dc, Item.PLATFORM, platformId))
			{
				errors.add("[Line "+lineNo+"] Unknown platform: "+platformId);
			}
			if (itemIsMissing(dc, Item.PLATFORMVARIANT, variantId))
			{
				errors.add("[Line "+lineNo+"] Unknown platform variant: "+variantId);
			}
			if (itemIsMissing(dc, Item.ITEMSUBTYPE, subtypeId))
			{
				errors.add("[Line "+lineNo+"] Unknown item subtype: "+subtypeId);
			}
			for (Map.Entry<Integer, Integer> entry : fileTypes.entrySet())
			{
				if (itemIsMissing(dc, Item.DATAFILETYPE, entry.getKey()))
				{
					errors.add("[Line "+lineNo+"] Unknown file type: "+entry.getKey());
				}
				if (itemIsMissing(dc, Item.DATAFILETYPE, entry.getValue()))
				{
					errors.add("[Line "+lineNo+"] Unknown file type: "+entry.getValue());
				}
			}
		}

		
		private boolean itemIsMissing(DbControl dc, Item type, Integer id)
		{
			if (id != null)
			{
				try
				{
					type.getById(dc, id);
				}
				catch (ItemNotFoundException ex)
				{
					return true;
				}
			}
			return false;
		}
		
		boolean matches(Rba rba)
		{
			boolean matched = true;
			if (rawDataType != null)
			{
				matched &= rawDataType.equals(rba.rawDataType);
			}
			if (platformId != null)
			{
				matched &= platformId.equals(rba.platformId);
			}
			if (variantId != null)
			{
				matched &= variantId.equals(rba.variantId);
			}
			return matched;
		}
		
		boolean matches(Itemlist list)
		{
			boolean matched = true;
			if (rawDataType != null)
			{
				matched &= rawDataType.equals(list.rawDataType);
			}
			return matched;
		}
	}
	
	/**
		Used for adding log entries to the database if that is enabled.
		For each migrated raw bioassay it will create two UPDATE entries
		one for the raw bioassay and one for the derived bioassay.
	*/
	public static class MigratedLogger
		implements EntityLogger
	{

		public MigratedLogger()
		{}
		
		@Override
		public void logChanges(LogManager logManager, EntityDetails details) 
		{
			LogControl logControl = logManager.getLogControl();
			Rba entry = (Rba)details.getEntity();
			
			ChangeHistoryDetailData change = new ChangeHistoryDetailData();
			change.setChangeType(ChangeType.UPDATE.getValue());
			change.setItemId(entry.rbaId);
			change.setItemType(Item.RAWBIOASSAY.getValue());
			change.setChangeInfo("Migrated to derived bioassay");
			((DbLogManager)logManager).logChangeDetails(change, details);

			change = new ChangeHistoryDetailData();
			change.setChangeType(ChangeType.UPDATE.getValue());
			change.setItemId(entry.dbaId);
			change.setItemType(Item.DERIVEDBIOASSAY.getValue());
			change.setChangeInfo("Migrated from raw bioassay");
			((DbLogManager)logManager).logChangeDetails(change, details);
		}
	}
}
