/**
	$Id: BooleanFormatter.java 8225 2024-06-14 06:38:29Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.formatter;

import net.sf.basedb.util.Values;
import net.sf.basedb.util.excel.ExcelFormatter;
import net.sf.basedb.util.excel.ExcelValue;

/**
	Formats a boolean value. It is possible to select between yes/no, true/false, and 1/0
	mappings. 

	@author nicklas
	@version 2.0
	@base.modified $Date: 2024-06-14 08:38:29 +0200 (Fri, 14 Jun 2024) $
*/
public class BooleanFormatter
	implements Formatter<Boolean>, ExcelFormatter<Boolean, Object>
{

	/**
		Create a formatter for boolean values.
		@return A formatter or null if there is no type
		@since 3.19.12
	*/
	public static BooleanFormatter fromType(String type)
	{
		Type t = Type.getByName(type);
		return t == null ? null : new BooleanFormatter(t);
	}
	
	private final Type type;

	/**
		Create a new boolean formatter.
		@param type The type of output
	*/
	public BooleanFormatter(Type type)
	{
		this.type = type;
	}
	
	/*
		From the Formatter interface
		-------------------------------------------
	*/
	@Override
	public String format(Boolean value)
	{
		return value == null ? "" : type.getString(value);
	}
	@Override
	public Boolean parseString(String value)
	{
		return Values.getBoolean(value);
	}
	// -------------------------------------------
	
	/**
		@since 3.15
	*/
	@Override
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public ExcelValue toExcelValue(Boolean value) 
	{
		return type.getExcelValue(value);
	}

	/**
		Enum for holding the various types of boolean output.
	*/
	public enum Type
	{
		/**
			Display yes/no.
		*/
		YES_NO("yes", "no"),

		/**
			Display true/false.
		*/
		TRUE_FALSE("true", "false"),

		/**
			Display TRUE/FALSE
			@since 3.19.12
		*/
		TRUE_FALSE_CAPS("TRUE", "FALSE"),
		
		/**
			Display 1/0.
		*/
		ONE_ZERO("1", "0")
		{
			@Override
			public ExcelValue<?> getExcelValue(Boolean value) 
			{
				return  ExcelValue.asInt(value == null ? null : (Boolean.TRUE.equals(value) ? 1 : 0));
			}
		};
		
		/**
			Get the type by name of the static constant defined in this class.
			
			@param cName The name of the static constant
			@return A subtype object or null if not found
			@since 3.19.12
		*/
		public static Type getByName(String cName)
		{
			for (Type t : Type.values())
			{
				if (t.name().equals(cName)) return t;
			}
			return null;
		}

		
		private final String trueValue;
		private final String falseValue;
		
		Type(String trueValue, String falseValue)
		{
			this.trueValue = trueValue;
			this.falseValue = falseValue;
		}
		
		public String getString(boolean value)
		{
			return value ? trueValue : falseValue;
		}
		
		public ExcelValue<?> getExcelValue(Boolean value)
		{
			return ExcelValue.asBoolean(value);
		}
	}

}
